from scipy.special import digamma as _digamma, betainc as _betainc, beta as _beta
import numpy as _np
import zentropy.estimators.utils as _utils

def surprisal_from_counts(counts,base=2,p_estimate=0,eps=1e-8):
    N = counts.sum()
    bias_correction = _betainc(N+1,eps,p_estimate)*_beta(N+1,eps)
    return (_digamma(N+1) - _digamma(counts+1) + bias_correction)/_np.log(base)

def weighted_surprisal_from_counts(counts,base=2,p_estimate=0,eps=1e-8):
    N = counts.sum()
    n_psi_ns = _np.empty_like(counts)
    n_psi_ns[counts==0] = 0
    n_psi_ns[counts>0] = counts*_digamma(counts)
    bias_correction = _betainc(N,eps,p_estimate)*_beta(N,eps)
    return ((_digamma(N) + bias_correction)*counts - n_psi_ns)/(N*_np.log(base))

def entropy_from_counts(p_counts,q_counts=None,base=2,**surprisal_kwargs):
    if q_counts is None:
        return weighted_surprisal_from_counts(p_counts,base=base,p_estimate=0,**surprisal_kwargs).sum()
    else:
        return (p_counts*surprisal_from_counts(q_counts,base=base,**surprisal_kwargs)).sum()/p_counts.sum()



from sklearn.neighbors import NearestNeighbors as _NearestNeighbors, KDTree as _KDTree
import numpy as _np

def get_knn_distance(xs,k=3,metric="chebyshev",**knn_kwargs):
    nn = _NearestNeighbors(metric=metric,n_neighbors=k,**knn_kwargs)
    nn.fit(xs)
    radii = _np.nextafter(nn.kneighbors()[0][:,-1],0)
    return radii

def count_proper_neighbors(xs,radii,ys=None,metric="chebyshev",**kdtree_kwargs):
    sub = 0
    if ys is None:
        ys = xs
        sub = 1

    kd = _KDTree(ys,metric=metric,**kdtree_kwargs)
    return kd.query_radius(xs,radii,return_distance=False,count_only=True) - sub

def get_neighbors(xs,radii,ys=None,metric="chebyshev",**kdtree_kwargs):
    if ys is None:
        ys = xs

    kd = _KDTree(ys,metric=metric,**kdtree_kwargs)
    return kd.query_radius(xs,radii,return_distance=False,)






from scipy.special import digamma as _digamma
import numpy as _np
import zentropy.knn_utils as _knn_utils

def surprisal_from_points(xs,ys=None,k=3,r=None,base=2,):
    if r is None:
        r = _knn_utils.get_knn_distance(xs,k,)
    elif hasattr(r,"__len__"):
        if len(r)!=xs.shape[0]:
            raise ValueError("Number of radii must be equal to number of samples if radius is not constant")
    else:
        r = _np.array([r]*xs.shape[0])

    n_neighbors = _knn_utils.count_proper_neighbors(xs,r,ys=ys,)
    return (_digamma(xs.shape[0]) - _digamma(n_neighbors) + xs.shape[1]*_np.log(r))/_np.log(base)

def entropy_from_points(xs,ys=None,k=3,r=None,base=2,):
    surprisals = surprisal_from_points(xs,ys=ys,k=k,r=r,base=base)
    return surprisals.mean()
        


from setuptools import setup, Extension
from Cython.Build import cythonize
import numpy
extensions = [
    Extension("zentropy.lib", ["lib/knnpermute.c",],
        include_dirs=[numpy.get_include(),]),
]
setup_args = dict(
    ext_modules = extensions,
    package_data = {"lib":["knnpermute.c"]}
)
setup(**setup_args)


import numpy as _np
import zentropy.knn_utils as _knn_utils
import zentropy.estimators.utils as _utils
import zentropy.estimators.continuous as _continuous
import zentropy.estimators.discrete as _discrete

def entropy(data,discrete_vars=[],cvars=[],base=2,k=3,r=None):
    data, df_discrete_vars = _utils.check_dataframe(data)
    if df_discrete_vars is not None:
        discrete_vars = df_discrete_vars
    cont_vars = [c for c in data.columns if not(c in discrete_vars)]
    if discrete_vars and cont_vars:
        groupings = data.groupby(discrete_vars)
        surprisals = groupings.apply(
            lambda df: _continuous.entropy_from_points(
                df[cont_vars],
                k=k,
                r=(r[df.index] if hasattr(r,"__len__") else r),
                base=base
            )
        )
        counts = data.value_counts(discrete_vars)
        return _discrete.entropy_from_counts(counts,base=2) + (surprisals*counts).sum()/counts.sum()
    elif discrete_vars:
        return _discrete.entropy_from_counts(data.value_counts(discrete_vars),base=base)
    else:
        return _continuous.entropy_from_points(data,k=k,r=r,base=base)

def mutual_info(data,*mutual_vars,cvars=[],discrete_vars=[],base=2,k=3,r=None):
    data, df_discrete_vars = _utils.check_dataframe(data)
    if df_discrete_vars is not None:
        discrete_vars = df_discrete_vars
    cont_vars = [c for c in data.columns if not(c in discrete_vars)]
    all_vars = list(mutual_vars)+cvars

    if cont_vars:
        if discrete_vars and r is None:
            groupings = data.groupby(discrete_vars)
            indices = groupings.apply(lambda df:df.index)
            r_groups = groupings.apply(
                lambda df: _knn_utils.get_knn_distance(df[cont_vars],k=k,)
            )
            r = _np.zeros([data.shape[0]])
            for g in r_groups.index:
                r[indices[g]] = r_groups[g]
        elif r is None:
            r = _knn_utils.get_knn_distance(data,k=k,)
    
    all_vars = _np.array(list(mutual_vars)+cvars,dtype=object)
    n_vars = len(all_vars)
    combo_index = _utils.combination_index(mutual_vars,all_vars)
    coefficients = _utils.entropy_to_atoms_matrix(n_vars)[combo_index]
    pos_coefficients = (coefficients != 0)
    entropy_vars = [list(all_vars[a]) for a in _utils.atoms_to_variables_matrix(n_vars)[pos_coefficients].astype(bool)]
    entropies = _np.array([entropy(data[_utils.reduce_to_list(varlist)],base=base,r=r) for varlist in entropy_vars])
    return entropies.dot(coefficients[pos_coefficients])



import numpy as _np
import pandas as _pd
import itertools as _it

def check_dataframe(xs,):
    if (type(xs) is _pd.Series):
        df = xs.to_frame()
        return df,list(df.select_dtypes(exclude=_np.inexact).columns)
    elif (type(xs) is _pd.DataFrame):
        return xs, list(xs.select_dtypes(exclude=_np.inexact).columns)
    else:
        return _pd.DataFrame(xs), None
    
def atoms_to_entropy_matrix(n_vars):
    combos = atoms_to_variables_matrix(n_vars)
    atoms_to_entropy = (combos@combos.T>0).astype(int)    
    return atoms_to_entropy

def atoms_to_variables_matrix(n_vars):
    return _np.array([list(reversed([int(x) for x in bin(k)[2:]]))+[0]*(n_vars-1-int(_np.floor(_np.log(k)/_np.log(2)))) for k in range(1,2**n_vars)])

def entropy_to_atoms_matrix(n_vars):
    return _np.linalg.inv(atoms_to_entropy_matrix(n_vars)).astype(int)

def combination_index(some_vars,all_vars):
    return int("".join(reversed([f"{i}" for i in (_np.isin(_np.array(all_vars),some_vars).astype(int))])),2)-1

def reduce_to_list(some_vars):
    return [var for v in some_vars for var in (v if (hasattr(v,"__len__") and (type(v) is not str)) else [v])]